import jDisco.*;
import jDisco.Process;

/*
 * Simulation of an aircraft arresting gear.
 * 
 * The aims of the simulation is to determine the range of aircraft weights 
 * and speeds that can be accommodated without exceeding the woking limits 
 * of the cabels or the allowable piston travel.
 *
 * The model has been taken from
 *
 *     Advanced Continuous Simulation Language (ACSL)
 *    Reference Manual, 1999, (p. A-34 - A-40)
 *     http://www.ACSL.sim.com
 */

public class AircraftArrestingGear extends Process {
    static final double m1 = 140.0, m2 = 45.28, m3 = 20.0;
    static final double k1 = 4550.0, k2 = 25300.0;
    static final double h = 125.0, xmax = 1000.0;    
    
    Variable x, xd, y2, y2d, y3, y3d;
    double y1, sth;
    Table fy3;
    double speed;
    Switcher sw1, sw2;
    
    public AircraftArrestingGear(double speed) {
        this.speed = speed;
    }
    
    public void derivatives() {
        y1 = Math.sqrt(Math.pow(x.state, 2) + Math.pow(h, 2)) - h;
        sth = x.state / (h + y1);
        double fk1 = k1 * (sw1.isOn() ? y1 - 2 * y2.state : 0);
        double fk2 = k2 * (sw2.isOn() ? y2.state - y3.state : 0);
        double fd  = fy3.value(y3.state) * Math.pow(y3d.state, 2);
        y3d.rate = (fk2 - fd) / m3; 
        y2d.rate = (2.0 * fk1 - fk2) / m2;
        xd.rate  =  -2.0 * fk1 * sth / m1;
        y3.rate  = y3d.state;
        y2.rate  = y2d.state;
        x.rate  = xd.state;
    }
    
    public class Switcher extends Process {
        public Switcher(final Condition cond) {
            this.cond = cond;
            not_cond = new Condition() {
                public boolean test() {
                    return !cond.test();
                }
            };
        }
        
        public void actions() {
            while (true) {
                on = false;
                 waitUntil(cond);
                 on = true;
                 waitUntil(not_cond); 
            }
        }
        
        public boolean isOn() {
            return on;
        }
        
        private Condition cond, not_cond;
        private boolean on;
    }
    
    public void actions() {
        x   = new Variable(0.0).start();
        xd  = new Variable(speed).start();
        y2  = new Variable(0.0).start();
        y2d = new Variable(0.0).start();
        y3  = new Variable(0.0).start();
        y3d = new Variable(0.0).start();
        fy3 = new Table();
        fy3.add(-10,  8.33); fy3.add(  0, 8.33); fy3.add( 30,  4.0); 
        fy3.add( 60,  1.6);  fy3.add(120, 5.2);  fy3.add(150,  5.2); 
        fy3.add(180,  6.6);  fy3.add(210, 8.3);  fy3.add(240, 10.7); 
        fy3.add(270, 16.0);  fy3.add(282, 21.0); fy3.add(294, 28.0); 
        fy3.add(306, 41.0);  fy3.add(312, 50.0); fy3.add(324, 90.0); 
        fy3.add(350, 90.0);
        sw1 = new Switcher(new Condition() {
            public boolean test() {
                return y1 > 2 * y2.state;
            }
        });
        sw2 = new Switcher(new Condition() {
            public boolean test() {
                return y2.state > y3.state;
            }
        });
        activate(sw1);
        activate(sw2);
        start();
        final Graph graph_x   = new Graph(" x (speed = " + speed + ") ");
        final Graph graph_xd  = new Graph(" xd (speed = " + speed + ") ");
        final Graph graph_xdd = new Graph(" xdd (speed = " + speed + ") ");
         class StateReporter extends Reporter {    
            public void actions() {
                graph_x.add(x.state);
                graph_xd.add(xd.state);
                graph_xdd.add(xd.rate);
            }
        }
        new StateReporter().setFrequency(0.2).start();
        waitUntil(new Condition() {
            public boolean test() {
                return time() >= 10 || x.state > xmax;
            }
        });
        graph_x.show(); 
        graph_xd.show(); 
        graph_xdd.show(); 
    }
    
    public static void main(String[] args) {
        activate(new AircraftArrestingGear(200.0));
        activate(new AircraftArrestingGear(290.0));
    }
}
