import jDisco.*;
import jDisco.Process;

/*
 * Numerical solution of Burgers' partial differential equation:
 *
 *    du/dt = -u * du/dt + 0.003 * d^2u/dx^2
 *
 * on the interval [0,1] and for 0 < t <= 1.1. 
 * 
 * Initial and boundary conditions:
 *
 *    u(0,x) = p(0,x), u(t, 0) = p(t,0), u(t,1) = p(t,1)
 *
 * where
 *
 *    p(t,x) = (0.1*exp(-a) + 0.5*exp(-b) + exp(-c)) / 
 *             (exp(-a) + exp(-b) + exp(-c))
 * 
 * and
 *
 *    a = 50/3(x - 0.5 + 4.95t) 
 *    b = 950/2(x - 0.5 + 0.75t)
 *    c = 500/2(x - 0.375)
 *
 * Reference:
 *    Sincovec, R.F., Madsen, N. K.
 *    Software for Nonlinear Partial Differential Equations
 *    ACM Trans. Math. Softw., Vol. 1, No. 3, 1975 (pp. 232 - 260).    
 */ 

public class Burger extends Process {
    PDEVariable u;
        
    class U extends PDEVariable {
        U(double xLeft, double xRight, int nPoints) {
            super(xLeft, xRight, nPoints);
            c = 1;
        }
        
        public double rate() { return -state() * dx() + 0.003 * d2x(); }
        
        public double initialState() { return trusol(0, x()); }
        
        public double b1Left()  { return 1; }
        public double b3Left()  { return trusol(time(), 0); }
        public double b1Right() { return 1; }
        public double b3Right() { return trusol(time(), 1); }
        
        double trusol(double t, double x) {
            double anu = 3.0e-3;
            double a = -(x - 0.5 + 4.95 * t) / (20 * anu);
            double b = -(x - 0.5 + 0.75 * t) / (4 * anu);
            double c = -(x - 0.375) / (2 * anu);
            a = Math.exp(a);
            b = Math.exp(b);
            c = Math.exp(c);
            return (0.1 * a + + 0.5 * b + c) / (a + b + c);
        }
    } 
    
     public void actions() {
         dtMin = 1.0e-5; dtMax = 0.1; 
         maxAbsError = maxRelError = 1.0e-5;
         u = new U(0, 1, 401);
         u.start();
         hold(1.1);
         u.show();
    }
        
    public static void main(String[] args) {
        activate(new Burger());
    }
}


