import jDisco.*;
import jDisco.Process;

/*
 * This example program demonstrates the use of jDisco in a combined 
 * simulation involving partial differential equations. 
 * 
 * The example has to do with the central heating of a building. The model 
 * has been taken from 
 *
 *     Cellier, F.E. and Blitz, A.E.
 *     GASP-V: A universal simulation package 
 *     Simulation of Systems, L. Dekker (editor)
 *     North-Holland, 1976, pp. 391-402.
 *
 * The building is modelled by a stick of length one. The left side of the 
 * stick represents the centre of the building where the heating takes place. 
 * The right side represents the walls of the building.  
 *      
 * The temperature distribution of the building is modelled using the following 
 * heat diffusion equation: 
 *
 *    dU/dt = 0.5 * d2u/dx2
 *                    
 * The heating of the central room is modelled by the equation (left boundary 
 * condition)
 *
 *    U = 30Z
 *
 * where Z is described by the ordinary first order differential equation 
 *
 *    dZ/dt = 4(1-Z)        , if the heating is on
 *                   
 *    dZ/dt = -4Z           , if the heating is off.  
 *
 * At night time (between 7 p.m. and 7 a.m.) the heating is always off. 
 * During the day the heating is on when the wall temperature falls below 
 * 19.5 degrees centigrade, and is turned off as soon as the wall temperature 
 * goes above 22.5 degrees centigrade. 
 *
 * The simulation starts at 6 a.m. with zero degrees centigrade throughout 
 * the whole building. 
 *
 * The model is far away from physical reality, but has been chosen intentionally 
 * since it demonstrates the use of partial differential equations, 
 * ordinary differential equations, time-events, as well as state-events in one 
 * single model.
 */

public class CentralHeatingSystem extends Process {
    Variable z;
    Temperature temperature;
    boolean heatingOn, dayTime;
    
    class zEquation extends Continuous {
        public void derivatives() {
            z.rate = heatingOn ? 4 * (1 - z.state) : -4 * z.state;
        }
    }
        
    class Temperature extends PDEVariable {
        Temperature(double xLeft, double xRight, int nPoints) {
            super(xLeft, xRight, nPoints);
        }
        
        public double rate() { return 0.5 * d2x(); }
        
        public double b1Left() { return 1; }
        public double b3Left() { return 30 * z.state; }
        public double b2Right() { return 1; }
    } 
    
    class Cycle extends Process {
        public void actions() {
            while (true) {
                dayTime = true;
                hold(12);
                dayTime = false;
                hold(12);
            }
        }
    }
    
    class Thermostat extends Process {
        public void actions() {
            while (true) {
                waitUntil(new Condition() {
                    public boolean test() {
                        return dayTime;
                    }
                });
                heatingOn = true;
                final Variable t = (Variable) temperature.points().last();
                waitUntil(new Condition() {
                    public boolean test() {
                        return !dayTime || t.state > 22.5;
                    }
                });
                heatingOn = false;
                waitUntil(new Condition() {
                    public boolean test() {
                        return !dayTime || t.state < 19.5;
                    }
                }); 
            }
        }
    }
    
    Graph a = new Graph(" a ");
    Graph b = new Graph(" b ");
    
     public void actions() {
         dtMin = 1.0e-5; dtMax =0.251; maxAbsError = maxRelError = 0.01;
         z = new Variable(0).start();
         new zEquation().start();
         temperature = new Temperature(0, 1, 11);
         temperature.start();
         hold(1);
         activate(new Cycle());
         activate(new Thermostat());
         new StateReporter().start();
         hold(15.5);
         a.show();
         b.show();
    }
    
    class StateReporter extends Reporter {    
        public void actions() {
            PDEVariable.Point p = (PDEVariable.Point) temperature.points().first();
            a.add(p);
            p = (PDEVariable.Point) temperature.points().last();
            b.add(p);
        }
    }
        
    public static void main(String[] args) {
        activate(new CentralHeatingSystem());
    }
}


