import jDisco.*;
import jDisco.Process;

/*
 * PHYSBE is an acronym for Physiological Benchmark Simulation Experiment, 
 * and is a benchmark problem which has been widely used in comparing 
 * continuous simulation languages. It is a mathematical model of the human 
 * cardiovascular system and relates pressure, volume and flow.
 * 
 * This implementation cannot be recommended! Not only is it "unstructured", 
 * but, what is worse, numerical problems may arise because discrete changes 
 * are executed directly by a continuous process (the use of the max-method 
 * makes the second-order derivatives of the volumes discontinuous). 
 *  
 * References:
 *
 *     Benham, R.D. 
 *     An ISL-8 and ISL-15 study of the physiological simulation 
 *     benchmark experiment
 *     SIMULATION, Vol. 18, no. 4, March 1972, pp. 152-156.
 * 
 *     Advanced Continuous Simulation Language (ACSL)
 *     Reference Manual, 1999, (p. A-13 - A-19)
 *     http://www.ACSL.sim.com  
 */

public class PHYSBE_2 extends Process {
    double startTime, 
           pap, pvp, plv, pao, psc, pvc, prv,
           fpv, fps, fmv, fav, fas, fvs, ftv;
    Variable vap, vvp, vlv, vao, vsc, vvc, vrv;
    Table acrv, aclv;
    
    public void derivatives() {
        pap = 0.133 * vap.state;
        pvp = 0.033 * vvp.state;
        pao = 0.8 * vao.state; 
        psc = 0.0153 * vsc.state;  
        pvc = 0.004 * vvc.state;
        plv = aclv.value(time() - startTime) * vlv.state;
        prv = acrv.value(time() - startTime) * vrv.state;
        
        fpv = 90 * Math.max(prv - pap, 0);
        fps = 7 * (pap - pvp);
        fmv = 17 * Math.max(pvp - plv, 0);
        fav = 80 * Math.max(plv - pao, 0);
        fas = 1.63 * (pao - psc);
        fvs = 1.65 * (psc - pvc);
        ftv = 78 * Math.max(pvc - prv, 0);

        vap.rate = fpv - fps;
        vvp.rate = fps - fmv;
        vlv.rate = fmv - fav; 
        vao.rate = fav - fas;
        vsc.rate = fas - fvs;
        vvc.rate = fvs - ftv;
        vrv.rate = ftv - fpv; 
    }
    
    class HeartBeat extends Process {
        HeartBeat(double period) {
            this.period = period;
        }
        
        public void actions() {
            while (true) {
                startTime = time();
                hold(period);
            }
        }
        
        double period;
    }
    
     public void actions() {
         vap = new Variable( 120).start();
         vvp = new Variable( 240).start();
         vvc = new Variable( 500).start();
         vao = new Variable( 100).start();
         vsc = new Variable(3340).start();
         vlv = new Variable( 150).start();
         vrv = new Variable( 150).start();
         start();
         
         Table t = aclv = new Table();
         t.add(0,  0.0066); t.add(0.04, 0.60); t.add(0.08, 1.00); t.add(0.12, 1.25);
         t.add(0.16, 1.40); t.add(0.20, 1.50); t.add(0.24, 1.60); t.add(0.28, 1.60);
         t.add(0.32, 1.50); t.add(0.36, 1.00); t.add(0.40, 0.0066); 
         t = acrv = new Table();
         t.add(0,  0.0033); t.add(0.04, 0.05); t.add(0.08, 0.10); t.add(0.12, 0.15);
         t.add(0.16, 0.19); t.add(0.20, 0.24); t.add(0.24, 0.30); t.add(0.28, 0.36);
         t.add(0.32, 0.40); t.add(0.36, 0.30); t.add(0.40, 0.0033); 
         
         final Graph graph_vrv = new Graph(" Right Ventricle Volume ");
         final Graph graph_pvp = new Graph(" Venuous Pulmonary Pressure ");
         final Graph graph_pao = new Graph(" Aorta Pressure ");
         new Reporter() {    
            public void actions() {
                graph_vrv.add(vrv);
                graph_pvp.add(pvp);
                graph_pao.add(pao);
            }
        }.setFrequency(0.02).start();
        activate(new HeartBeat(1));
        hold(2);
        graph_vrv.show();
        graph_pvp.show();
        graph_pao.show();
    }
    
    public static void main(String[] args) {
        activate(new PHYSBE_2());
    }
}


