import jDisco.*;
import jDisco.Process;

/*
 * Simulation of a chemical reaction process.
 *
 * The reactor system consists of a compressor, a surge tank, and four reactors. 
 * The reactors are charged with reactants, which are supplied with hydrogen from
 * the compressor through the surge tank. The reactants react under pressure with 
 * the hydrogen, which causes the concentration of the reactants to decrease. 
 * The effective pressure in each reactor is automatically adjusted to the minimum 
 * of the surge tank pressure and critical pressure (100 psia). 
 *
 * When the concentration of a reactant decreases to 10% of its initial value, 
 * the reaction is considered complete and the reactor is turned off and made 
 * ready for a fresh batch of reactant.
 *
 * Initially the surge tank pressure is 500 psia. If the pressure falls below the 
 * critical value of 100 psia, the last reactor started will be turned off immediately. 
 * The other reactors will continuous, but no reactor will be started as long as surge 
 * tank pressure is below a nominal pressure of 150 psia. 
 *
 * If two or more reactors can start at the same time, the reactor with the highest 
 * value of accumulated batch processing time will be started first.
 *
 * The reactor system involves both discrete events and continuous state variables. 
 * The starting and stopping of reactors are discrete events. Between these events 
 * concentration and pressure vary continuously.
 *
 * The continuous processes of the system are described in the class Reactions. 
 * The class is used to define a single continuous process having a variable number 
 * of differential equations. The process computes the active reactors' rate of 
 * change in surge tank pressure.
 *
 * The discrete processes of the system are described in the classes Reactor and 
 * Interrupter. 
 *
 * Class Reactor describes the reactors and associated events. Starting a reactor 
 * is allowed only if the surge tank pressure is above the nominal pressure. 
 * When the reaction process is completed, the reactor is stopped, cleaned and 
 * recharged.
 *
 * Class Interrupter describes the discrete events occurring whenever surge tank 
 * pressure drops to the critical pressure. An Interrupter-object sees that the 
 * last reactor started is turned off when the pressure falls below the critical 
 * pressure. Further, the object ensures that the effective pressure in the 
 * reactors is the minimum of surge tank pressure and the critical pressure.
 */

public class ReactorSystem extends Process {
    static final double flowFromCompressor = 4.3523;
    static final double factor = 107.03;
    static final double nominalPressure = 150;
    static final double criticalPressure = 100;    
    Reactor[] reactor = { new Reactor(0.03466, 10, 0.1),
                          new Reactor(0.00866, 15, 0.4),
                          new Reactor(0.01155, 20, 0.2),
                          new Reactor(0.00770, 25, 0.5) };
    Variable pressure;
    double effectivePressure;
    Head started = new Head(); 
    Graph[] concentrationGraph = { new Graph("Concentration 1"), 
                                   new Graph("Concentration 2"), 
                                   new Graph("Concentration 3"), 
                                   new Graph("Concentration 4") };
    Graph pressureGraph = new Graph("Pressure");
    
    public void actions() {
        dtMin = 0.001; dtMax = 1; maxRelError = maxAbsError = 0.00001;
        pressure = new Variable(500).start();
        activate(reactor[0], at, 0.0);
        activate(reactor[1], at, 0.5);
        activate(reactor[2], at, 1.0);
        activate(reactor[3], at, 1.5);
        activate(new Interrupter());
        new Reactions().start();
        Reporter reporter = new StateReporter().setFrequency(0.0).start();
        hold(10);
        for (int i = 0; i < 4; i++)
            concentrationGraph[i].show();
        pressureGraph.show();
    }
    
    class Reactor extends Process {
        Reactor(double reactionConstant, double volume, double initialConcentration) {
            this.reactionConstant = reactionConstant;
            this.volume = volume;
            this.initialConcentration = initialConcentration;
            concentration = new Variable(initialConcentration);
        } 
        
        public void actions() {
            Random rand = new Random(7913);
            while (true) {
                double processingTime = 0;
                while (concentration.state > 0.1 * initialConcentration) {
                    waitUntil(new Condition() {
                        public boolean test() { 
                            return pressure.state >= nominalPressure; 
                        }
                    }, processingTime);
                    concentration.start();
                    into(started); 
                    double startTime = time();
                    waitUntil(new Condition() {
                        public boolean test() { 
                            return concentration.state <= 0.1 * initialConcentration; 
                        }
                    });
                    concentration.stop();
                    out(); 
                    processingTime += (time() - startTime);
                }
                concentration.state = initialConcentration;
                hold(rand.negexp(1)); 
                hold(Math.min(rand.normal(1, 0.5), 2));
            }
        }
        
        Variable concentration;
        double reactionConstant, volume, initialConcentration;
    }
    
    class Reactions extends Continuous {
        public void derivatives() {
            double flowToReactors = 0;
            for (Reactor r = (Reactor) started.first(); r != null; r = (Reactor) r.suc()) {
                r.concentration.rate = -r.reactionConstant * r.concentration.state
                                                           * effectivePressure;
                flowToReactors += -r.concentration.rate * r.volume;
            }
            pressure.rate = factor * (flowFromCompressor - flowToReactors);
        } 
    }
    
    class Interrupter extends Process {
        public void actions() {
            while (true) {
                effectivePressure = criticalPressure;
                waitUntil(new Condition() {
                    public boolean test() { 
                        return pressure.state <= criticalPressure;
                    }
                });
                activate((Process) started.last());
                effectivePressure = pressure.state;
                waitUntil(new Condition() {
                    public boolean test() { 
                        return pressure.state > criticalPressure;
                    }
                });
            }
        }
    }
    
    class StateReporter extends Reporter {
        public void actions() {
            for (int i = 0; i < 4; i++)
                concentrationGraph[i].add(reactor[i].concentration);
            pressureGraph.add(pressure);
        }
    }
    
    public static void main(String[] args) {
        activate(new ReactorSystem());
    }
}
