import jDisco.*;
import jDisco.Process;

/* 
 * Solution of the collection of 20 restricted three-body test
 * problems presented in
 *
 *  Sharp, P. W.
 *  A collection of restricted three-body test problems.
 *  University of Auckland, Research Reports 460, 2001
 *  http://www.math.auckland.ac.nz/~sharp/papers/r3b_collection_I.pdf
 */
 
public class ThreeBodyProblem extends Process {
    String name;
    double mju, x0, y0, vx0, vy0, T, tolerance;
    Variable x, y, vx, vy;
    Graph graph;
    int N;
    double hSum;

    public ThreeBodyProblem(String name, double mju, double x0, double vx0,
                            double y0, double vy0,
                            double T, double tolerance) {
        this.name = name;
        this.mju = mju;
        this.x0 = x0;
        this.vx0 = vx0;
        this.y0 = y0;
        this.vy0 = vy0;
        this.T = T;
        this.tolerance = tolerance;
        graph = new Graph(name);
    }

    public void derivatives() {
        x.rate = vx.state;
        y.rate = vy.state;
        double x1 = -mju, x2 = 1 - mju;
        double ySquare = Math.pow(y.state, 2);
        double r13 = Math.pow(Math.sqrt(Math.pow(x.state - x1, 2) + ySquare), 3);
        double r23 = Math.pow(Math.sqrt(Math.pow(x.state - x2, 2) + ySquare), 3);
        vx.rate = x.state + 2 * y.rate - (1 - mju) * (x.state - x1) / r13
                  - mju * (x.state - x2) / r23;
        vy.rate = y.state - 2 * x.rate - (1 - mju) * y.state / r13
                - mju * y.state / r23;
    }

    public void actions() {
        dtMin = 1e-10;
        dtMax = 1;
        maxAbsError = maxRelError = tolerance;
        x = new Variable(x0).start();
        y = new Variable(y0).start();
        vx = new Variable(vx0).start();
        vy = new Variable(vy0).start();
        start();
        new Reporter() {
            public void actions() {
                N++;
                hSum += dt();
                graph.add(x, y);
            }
        }.setFrequency(0).start();
        long startTime = System.currentTimeMillis();
        hold(T);
        long stopTime = System.currentTimeMillis();
        graph.show();
        double sqSum = Math.pow(x.state - x0, 2)
                + Math.pow(y.state - y0, 2)
                + Math.pow(vx.state - vx0, 2)
                + Math.pow(vy.state - vy0, 2);
        double E = Math.sqrt(sqSum);
        if (name.length() < 3)
            System.out.print(" ");
        System.out.printf("%s: N = %4d, 10^3h = %4.1f, " +
                          "log10E = %.2f, time = %.3f s.\n",
                          name, N, 1000 * hSum / N, 
                          Math.log10(E), (stopTime - startTime) / 1000.0);
    }

    public static void main(String[] args) {
        double tolerance = 1E-10;

        activate(new ThreeBodyProblem("P1", 0.012277471, 0.994000, 0, 0,
                -0.21138987966945026683E+1,
                0.54367954392601899690E+1, tolerance));
        activate(new ThreeBodyProblem("P2", 0.012277471, 0.994000, 0, 0,
                -0.20317326295573368357E+1,
                0.11124340337266085135E+2, tolerance));
        activate(new ThreeBodyProblem("P3", 0.012277471, 0.994000, 0, 0,
                -0.20015851063790825224E+1,
                0.17065216560157962559E+2, tolerance));
        activate(new ThreeBodyProblem("P4", 0.012277471, 0.997000, 0, 0,
                -0.16251217072210773125E+1,
                0.2292972342344296948E+2, tolerance));
        activate(new ThreeBodyProblem("P5", 0.012277471, 0.879962, 0, 0,
                -0.66647197988564140807E+0,
                0.63006757422352314657E+1, tolerance));
        activate(new ThreeBodyProblem("P6", 0.012277471, 0.879962, 0, 0,
                -0.43965281709207999128E+0,
                0.12729711961022426544E+2, tolerance));
        activate(new ThreeBodyProblem("P7", 0.012277471, 0.879962, 0, 0,
                -0.38089067106386964470E+0,
                0.19138746281183026809E+2, tolerance));       
        activate(new ThreeBodyProblem("P8", 0.012277471, 0.997000, 0, 0,
                -0.18445010489730401177E+1,
                0.12353901248612092736E+2, tolerance));
        activate(new ThreeBodyProblem("P9", 0.012277471, 1.00000, 0, 0,
                -0.16018768253456252603E+1,
                0.1229438779669502330E+2, tolerance));
        activate(new ThreeBodyProblem("P10", 0.012277471, 1.00300, 0, 0,
                -0.14465123738451062297E+1,
                0.12267904265603897140E+2, tolerance));
        activate(new ThreeBodyProblem("P11", 0.012277471, 1.20000, 0, 0,
                -0.71407169828407848921E+0,
                0.18337451820715063383E+2, tolerance));
        activate(new ThreeBodyProblem("P12", 0.012277471, 1.20000, 0, 0,
                -0.67985320356540547720E+0,
                0.30753758552146029263E+2, tolerance));
        activate(new ThreeBodyProblem("P13", 0.012277471, 1.20000, 0, 0,
                -0.67153130632829144331E+0,
                0.43214375227857454128E+2, tolerance));
        activate(new ThreeBodyProblem("P14", 0.012277471, 1.20000, 0, 0,
                -0.66998291305226832207E+0,
                0.55672334134347612727E+2, tolerance));
        activate(new ThreeBodyProblem("P15", 0.012277471, 1.20000, 0, 0,
                -0.66975741517271092087E+0,
                0.68127906604713772763E+2, tolerance));
        activate(new ThreeBodyProblem("P16", 0.000953875, -1.02745, 0, 0,
                0.40334488290490413053E-1,
                0.18371316400018903965E+3, tolerance));
        activate(new ThreeBodyProblem("P17", 0.000953875, -0.976680, 0, 0,
                -0.61191623926410837000E-1,
                0.1773324113152448300E3, tolerance));
        activate(new ThreeBodyProblem("P18", 0.000953875, -0.766650, 0, 0,
                -0.51230158665978820282E+0,
                0.17660722897242937108E+3, tolerance));
        activate(new ThreeBodyProblem("P19", 0.000953875, -1.09137, 0, 0,
                0.14301959822238380020E+0,
                0.82949461922342093092E+2, tolerance));
        activate(new ThreeBodyProblem("P20", 0.000953875, -1.10137, 0, 0,
                0.15354250908611454510E+0,
                0.60952121909407746612E+2, tolerance));
    }
}
