import jDisco.*;
import jDisco.Process;

/*
 * PHYSBE is an acronym for Physiological Benchmark Simulation Experiment, 
 * and is a benchmark problem which has been widely used in comparing 
 * continuous simulation languages. It is a mathematical model of the human 
 * cardiovascular system and relates pressure, volume and flow.
 * 
 * This version demonstrates the exploitation of Java's class concept for
 * "structured" modelling.
 *  
 * References:
 *
 *     Benham, R.D. 
 *     An ISL-8 and ISL-15 study of the physiological simulation 
 *     benchmark experiment
 *     SIMULATION, Vol. 18, no. 4, March 1972, pp. 152-156.
 * 
 *     Advanced Continuous Simulation Language (ACSL)
 *     Reference Manual, 1999, (p. A-13 - A-19)
 *     http://www.ACSL.sim.com  
 */

public class PHYSBE_1 extends Process {
    class Compartment {
        Compartment(Variable volume, double stiffness) {
            this.volume = volume;
            this.stiffness = stiffness;
            new PressureChange(this);
            new VolumeChange(this);
        }
        
        double stiffness() {
            return stiffness;
        }
        
        Variable volume;
        private double stiffness;
        double pressure;
        Connection input, output;
    }
    
    class HeartChamber extends Compartment {
        HeartChamber(Variable volume) {
            super(volume, 0);
            stiffness = new Table();
        }
        
        double stiffness() {
            return stiffness.value(time() - startTime);
        }
        
        Table stiffness;        
    }
    
    class PressureChange extends Continuous {
        PressureChange(Compartment c) {
            this.c = c;
            setPriority(-1);
            start();
        }
        
        public void derivatives() {
            c.pressure = c.stiffness() * c.volume.state;
        }
        
        Compartment c;
    }
    
    class Connection extends Continuous {
        Connection(Compartment from, Compartment to, double admittance) {
            this.from = from;
            this.to = to;
            this.admittance = admittance;
            setPriority(-2);
            start();
            from.output = to.input = this;
            if (from instanceof HeartChamber || to instanceof HeartChamber) 
                activate(new Valve(this), after, current());
        }
        
        public void derivatives() {
            flow = admittance * (from.pressure - to.pressure);
        }
        
        Compartment from, to;
        double admittance;
        double flow;
    }
    
    class VolumeChange extends Continuous {
        VolumeChange(Compartment c) {
            this.c = c;
            setPriority(-3);
            start();
            c.volume.start();
        }
        
        public void derivatives() {
            c.volume.rate = c.input.flow - c.output.flow;
        }
        
        Compartment c;
    }
    
    class Valve extends Process {
        Valve(Connection pipe) {
            this.pipe = pipe;
        }
        
        public void actions() {
            while (true) {
                waitUntil(new Condition() {
                    public boolean test() {
                        return pipe.from.pressure <= pipe.to.pressure;
                    }
                });
                pipe.stop();
                pipe.flow = 0;
                waitUntil(new Condition() {
                    public boolean test() {
                        return pipe.from.pressure > pipe.to.pressure;
                    }
                });
                pipe.start();
            }
        }
        
        Connection pipe;
    }
    
    class HeartBeat extends Process {
        HeartBeat(double period) {
            this.period = period;
        }
        
        public void actions() {
            while (true) {
                startTime = time();
                hold(period);
            }
        }
        
        double period;
    }
    
    Compartment arterialPulmonary, venousPulmonary,
                venaCava, aorta, systemicCirculation;
    HeartChamber leftVentricle, rightVentricle;    
    double startTime;
    
     public void actions() {
         arterialPulmonary     = new Compartment(new Variable( 120), 0.133);
         venousPulmonary       = new Compartment(new Variable( 240), 0.033);
         venaCava              = new Compartment(new Variable( 500), 0.004);
         aorta                 = new Compartment(new Variable( 100), 0.8);
         systemicCirculation   = new Compartment(new Variable(3340), 0.0153);
         leftVentricle         = new HeartChamber(new Variable(150));
         rightVentricle        = new HeartChamber(new Variable(150));
     
         new Connection(arterialPulmonary,   venousPulmonary,      7.00);
         new Connection(venousPulmonary,     leftVentricle,       17.00);
         new Connection(leftVentricle,       aorta,               80.00);
         new Connection(aorta,               systemicCirculation,  1.63);
         new Connection(systemicCirculation, venaCava,             1.65);
         new Connection(venaCava,            rightVentricle,      78.00);
         new Connection(rightVentricle,      arterialPulmonary,   90.00);
         
         Table t = leftVentricle.stiffness;
         t.add(0,  0.0066); t.add(0.04, 0.60); t.add(0.08, 1.00); t.add(0.12, 1.25);
         t.add(0.16, 1.40); t.add(0.20, 1.50); t.add(0.24, 1.60); t.add(0.28, 1.60);
         t.add(0.32, 1.50); t.add(0.36, 1.00); t.add(0.40, 0.0066); 
         t = rightVentricle.stiffness;
         t.add(0,  0.0033); t.add(0.04, 0.05); t.add(0.08, 0.10); t.add(0.12, 0.15);
         t.add(0.16, 0.19); t.add(0.20, 0.24); t.add(0.24, 0.30); t.add(0.28, 0.36);
         t.add(0.32, 0.40); t.add(0.36, 0.30); t.add(0.40, 0.0033); 
         
         final Graph graph_vrv = new Graph(" Right Ventricle Volume ");
         final Graph graph_pvp = new Graph(" Venuous Pulmonary Pressure ");
         final Graph graph_pao = new Graph(" Aorta Pressure ");
         new Reporter() {    
            public void actions() {
                graph_vrv.add(rightVentricle.volume);
                graph_pvp.add(venousPulmonary.pressure);
                graph_pao.add(aorta.pressure);
            }
        }.setFrequency(0.02).start();
        activate(new HeartBeat(1));
        hold(2);
        graph_vrv.show();
        graph_pvp.show();
        graph_pao.show();
    }
    
    public static void main(String[] args) {
        activate(new PHYSBE_1());
    }
}


